import numpy as np
import argparse
import torch
import os

from tqdm import tqdm
from PIL import Image

from model.buffer import DemoBuffer
from model.utils import set_seed
from model.env.env import make_env


def collect_demo(args):
    set_seed(args.seed)
    device = torch.device('cuda' if (torch.cuda.is_available() and not args.no_cuda) else 'cpu')
    env = make_env(env_id=args.env, device=device)
    env.compute_linearized_controller()

    buffer = DemoBuffer(
        buffer_size=args.steps,
        state_dim=env.n_dims,
        action_dim=env.n_controls,
        device=device
    )

    if not os.path.exists(args.path):
        os.mkdir(args.path)
    if not os.path.exists(os.path.join(args.path, args.env)):
        os.mkdir(os.path.join(args.path, args.env))

    epi_rewards = []
    cur_reward = 0
    cur_step = 0

    state = env.reset()
    for i in tqdm(range(args.steps)):
        action = env.noisy_u_nominal(state, std=args.std, bias=args.bias)
        next_state, reward, done, _ = env.step(action)
        cur_reward += reward
        buffer.append(state, action, reward, done, next_state)
        state = next_state
        cur_step += 1

        if args.render:
            data = env.render()
            im = Image.fromarray(data)
            im.save(os.path.join(args.path, args.env, f'step_{i}.png'))
            im.close()

        if done:
            tqdm.write(f'epi: {len(epi_rewards)}, reward: {cur_reward:.0f}, steps: {cur_step}')
            state = env.reset()
            epi_rewards.append(cur_reward)
            cur_reward = 0
            cur_step = 0

    buffer_name = f'size{args.steps}_std{args.std}_bias{args.bias}_reward{np.mean(epi_rewards):.0f}.pkl'
    buffer.save(os.path.join(args.path, args.env, buffer_name))
    print(f'buffer saved, mean reward: {np.mean(epi_rewards):.2f}, std: {np.std(epi_rewards):.2f}')


if __name__ == '__main__':
    parser = argparse.ArgumentParser()

    # custom
    parser.add_argument('--env', type=str, default='InvertedPendulum')
    parser.add_argument('--std', type=float, default=0.1)
    parser.add_argument('--bias', type=float, default=0.1)
    parser.add_argument('--render', action='store_true', default=False)

    # default
    parser.add_argument('--path', type=str, default='./demos')
    parser.add_argument('--seed', type=int, default=0)
    parser.add_argument('--no-cuda', action='store_true', default=False)
    parser.add_argument('--steps', type=int, default=20000)

    args = parser.parse_args()
    collect_demo(args)
