import torch

from torch import nn as nn
from torch.optim import Adam

from .base import Agent
from .utils import soft_update, disable_gradient
from ..env.base import ControlAffineSystem
from ..buffer import DemoBuffer
from ..network.policy import StateDependentPolicy
from ..network.value import TwinnedStateActionFunction


class SAC(Agent):
    """
    Implementation of SAC

    Reference:
    ----------
    [1] Haarnoja, T., Zhou, A., Abbeel, P., and Levine, S.
    Soft actor-critic: Off-policy maximum entropy deep reinforcement learning with a stochastic actor.
    In International Conference on Machine Learning, pp. 1861-1870. PMLR, 2018
    """

    def __init__(
            self,
            state_dim: int,
            action_dim: int,
            device: torch.device,
            goal_point: torch.Tensor,
            u_eq: torch.Tensor,
            state_std: torch.Tensor,
            ctrl_std: torch.Tensor,
            gamma: float = 0.99,
            batch_size: int = 256,
            rollout_length: int = 5 * 10**6,
            lr_actor: float = 3e-4,
            lr_critic: float = 3e-4,
            lr_alpha: float = 3e-4,
            units_actor: tuple = (128, 128),
            units_critic: tuple = (128, 128),
            start_steps: int = 10000,
            tau: float = 5e-3
    ):
        super().__init__(state_dim, action_dim, gamma, device)

        # replay buffer
        self.buffer = DemoBuffer(
            buffer_size=rollout_length,
            state_dim=state_dim,
            action_dim=action_dim,
            device=device
        )

        # actor
        self.actor = StateDependentPolicy(
            state_dim=state_dim,
            action_dim=action_dim,
            goal_point=goal_point,
            u_eq=u_eq,
            state_std=state_std,
            ctrl_std=ctrl_std,
            hidden_units=units_actor,
        ).to(device)

        # critic
        self.critic = TwinnedStateActionFunction(
            state_dim=state_dim,
            action_dim=action_dim,
            goal_point=goal_point,
            u_eq=u_eq,
            state_std=state_std,
            ctrl_std=ctrl_std,
            hidden_units=units_critic,
            hidden_activation=nn.Tanh()
        ).to(device)
        self.critic_target = TwinnedStateActionFunction(
            state_dim=state_dim,
            action_dim=action_dim,
            goal_point=goal_point,
            u_eq=u_eq,
            state_std=state_std,
            ctrl_std=ctrl_std,
            hidden_units=units_critic,
            hidden_activation=nn.Tanh()
        ).to(device).eval()

        soft_update(self.critic_target, self.critic, 1.0)
        disable_gradient(self.critic_target)

        # entropy coefficient
        self.alpha = 1.0
        # we optimize log(alpha) because alpha should be always bigger than 0
        self.log_alpha = torch.zeros(1, device=device, requires_grad=True)
        # target entropy is -|A|
        self.target_entropy = -float(action_dim)

        self.optim_actor = Adam(self.actor.parameters(), lr=lr_actor)
        self.optim_critic = Adam(self.critic.parameters(), lr=lr_critic)
        self.optim_alpha = torch.optim.Adam([self.log_alpha], lr=lr_alpha)

        self.batch_size = batch_size
        self.start_steps = start_steps
        self.tau = tau

    def act(self, x: torch.Tensor) -> torch.Tensor:
        if x.ndim == 1:
            x = x.unsqueeze(0)
        with torch.no_grad():
            return self.actor.forward(x)

    def save(self, path: str):
        torch.save(self.actor.state_dict(), path)

    def load(self, path: str, device: torch.device):
        self.actor.load_state_dict(torch.load(path, map_location=device))

    def set_controller(self, controller: nn.Module):
        self.actor.policy_net.load_state_dict(controller.state_dict())

    def is_update(self, step: int):
        return step >= max(self.start_steps, self.batch_size)

    def update(self):
        self.learning_steps += 1
        states, actions, rewards, dones, next_states = self.buffer.sample(self.batch_size)

        self.update_critic(states, actions, rewards, dones, next_states)
        self.update_actor(states)
        self.update_target()

    def step(self, env: ControlAffineSystem, state: torch.Tensor, t: int, step: int):
        t += 1

        if step <= self.start_steps:
            action = torch.tensor(env.action_space.sample(), device=self.device)
        else:
            with torch.no_grad():
                action, log_pi = self.actor.sample(state)

        next_state, reward, done, _ = env.step(action)
        mask = True if t == env.max_episode_steps else done

        self.buffer.append(state, action, reward, mask, next_state)

        if done or t == env.max_episode_steps:
            t = 0
            next_state = env.reset()

        return next_state, t

    def update_critic(
            self,
            states: torch.Tensor,
            actions: torch.Tensor,
            rewards: torch.Tensor,
            dones: torch.Tensor,
            next_states: torch.Tensor,
    ):
        """
        Update the critic for one step

        Parameters
        ----------
        states: torch.Tensor
            sampled states
        actions: torch.Tensor
            sampled actions according to the states
        rewards: torch.Tensor
            rewards of the s-a pairs
        dones: torch.Tensor
            whether is the end of the episode
        next_states: torch.Tensor
            next states give s-a pairs
        """
        curr_qs1, curr_qs2 = self.critic(states, actions)
        with torch.no_grad():
            next_actions, log_pis = self.actor.sample(next_states)
            next_qs1, next_qs2 = self.critic_target(next_states, next_actions)
            next_qs = torch.min(next_qs1, next_qs2) - self.alpha * log_pis
        target_qs = rewards + (1.0 - dones) * self.gamma * next_qs

        loss_critic1 = (curr_qs1 - target_qs).pow_(2).mean()
        loss_critic2 = (curr_qs2 - target_qs).pow_(2).mean()

        self.optim_critic.zero_grad()
        (loss_critic1 + loss_critic2).backward(retain_graph=False)
        self.optim_critic.step()

    def update_actor(self, states: torch.Tensor):
        """
        Update the actor for one step

        Parameters
        ----------
        states: torch.Tensor
            sampled states
        """
        actions, log_pis = self.actor.sample(states)
        qs1, qs2 = self.critic(states, actions)
        loss_actor = self.alpha * log_pis.mean() - torch.min(qs1, qs2).mean()

        self.optim_actor.zero_grad()
        loss_actor.backward(retain_graph=False)
        self.optim_actor.step()

        entropy = -log_pis.detach_().mean()
        loss_alpha = -self.log_alpha * (self.target_entropy - entropy)

        self.optim_alpha.zero_grad()
        loss_alpha.backward(retain_graph=False)
        self.optim_alpha.step()

        with torch.no_grad():
            self.alpha = self.log_alpha.exp().item()

    def update_target(self):
        """Update the critic target"""
        soft_update(self.critic_target, self.critic, self.tau)
